#ifndef _IRI_JOINTS_MODULE_H
#define _IRI_JOINTS_MODULE_H

// IRI ROS headers
#include <iri_ros_tools/module.h>
#include <iri_ros_tools/module_action.h>
#include <iri_ros_tools/module_service.h>
#include <iri_ros_tools/module_dyn_reconf.h>
#include <iri_ros_tools/watchdog.h>
#include <iri_ros_tools/timeout.h>
#include <iostream>

// ROS headers
#include <control_msgs/PointHeadAction.h>
#include <control_msgs/PointHeadActionGoal.h>
#include <control_msgs/FollowJointTrajectoryAction.h>
#include <sensor_msgs/JointState.h>
#include <geometry_msgs/PointStamped.h>
#include <iri_joints_msgs/pointHeadTrackerAction.h>

// Dynamic reconfigure header
#include <iri_joints_module/IriJointsModuleConfig.h>

//States
typedef enum {IRI_JOINTS_MODULE_IDLE, IRI_JOINTS_MODULE_POINT_WAIT, IRI_JOINTS_MODULE_TRACKING, IRI_JOINTS_MODULE_MOVE_ANGLES} iri_joints_module_state_t;

//Status
/**
  * \brief Possible module status returned by the get_status() function
  *
  * These are the possible status of the module, which are:
  *  * IRI_JOINTS_MODULE_RUNNING: indicates that the module is active and operating
  *                        properly.
  *  * IRI_JOINTS_MODULE_SUCCESS: indicates the last action goal has been reached
  *                        successfully.
  *  * IRI_JOINTS_MODULE_ACTION_START_FAIL: indicates that the goal could not be started.
  *  * IRI_JOINTS_MODULE_TIMEOUT: indicates that the goal did not complete in the
  *                        allowed time.
  *  * IRI_JOINTS_MODULE_FB_WATCHDOG: indicates that feedback has not been received for
  *                            a long time.
  *  * IRI_JOINTS_MODULE_ABORTED: indicates that the action goal could not be reached.
  *  * IRI_JOINTS_MODULE_PREEMPTED: indicates that the current action goal has been
  *                          canceled by another goal.
  *  * IRI_JOINTS_MODULE_REJECTED: indicates that the goal information is not valid and the
  *                         goal will not be executed.
  *  * IRI_JOINTS_MODULE_SET_PARAM_FAIL: indicates that it was impossible to set the value
  *                               of a parameter.
  *  * IRI_JOINTS_MODULE_PARAM_NOT_PRESENT: indicates that the parameter to be changed is
  *
  *  * IRI_JOINTS_MODULE_UNKNOWN_MAP: indicates that the desired map has not been found in
  *                            the database.
  *  * IRI_JOINTS_MODULE_INVALID_MODE: indicates that the navigation mode of the robot is
  *                             not set properly, so the navigation module will not
  *                             be operational.
  */
typedef enum {IRI_JOINTS_MODULE_RUNNING,
              IRI_JOINTS_MODULE_SUCCESS,
              IRI_JOINTS_MODULE_ACTION_START_FAIL,
              IRI_JOINTS_MODULE_TIMEOUT,
              IRI_JOINTS_MODULE_FB_WATCHDOG,
              IRI_JOINTS_MODULE_ABORTED,
              IRI_JOINTS_MODULE_PREEMPTED,
              IRI_JOINTS_MODULE_REJECTED,
              IRI_JOINTS_MODULE_SET_PARAM_FAIL,
              IRI_JOINTS_MODULE_PARAM_NOT_PRESENT,
              IRI_JOINTS_MODULE_NO_JOINT_STATES,
              IRI_JOINTS_MODULE_TARGET_LOST} iri_joints_module_status_t;

typedef enum {axis_x, axis_y, axis_z}Axis_enum_t;
/**
  * \brief Joints module
  *
  */
class CIriJointsModule : public CModule<iri_joints_module::IriJointsModuleConfig>
{
  private:
    /**
      * \brief
      *
      */
    iri_joints_module::IriJointsModuleConfig config;

    // PointHead action
    /**
      * \brief
      *
      */
    CModuleAction<control_msgs::PointHeadAction, iri_joints_module::IriJointsModuleConfig> point_to_action;
    /**
      * \brief
      *
      */
    control_msgs::PointHeadGoal point_to_goal;
    /**
      * \brief
      *
      */
    bool new_point_goal;
    bool cancel_point_to_pending;

    // dynamic reconfigure
    /**
      * \brief
      *
      */
    CModuleDynReconf<iri_joints_module::IriJointsModuleConfig> point_to_reconf;

    // JointStates
    /**
      * \brief
      *
      */
    CROSWatchdog joint_states_watchdog;
    /**
      * \brief
      *
      */
    ros::Subscriber joint_states_subscriber;
    /**
      * \brief
      *
      */
    void joint_states_callback(const sensor_msgs::JointState::ConstPtr& msg);

    //PointHead Tracker
    /**
      * \brief
      *
      */
    CModuleAction<iri_joints_msgs::pointHeadTrackerAction, iri_joints_module::IriJointsModuleConfig> track_action;
    /**
      * \brief
      *
      */
    iri_joints_msgs::pointHeadTrackerGoal track_goal;

    bool enable_tracker_pending;
    bool cancel_tracker_pending;
    ros::Publisher tracker_point_to_publisher_;
    control_msgs::PointHeadActionGoal target_goal;


    //FollowJointTrayectoryAction
    CModuleAction<control_msgs::FollowJointTrajectoryAction, iri_joints_module::IriJointsModuleConfig> move_to_angle_action;
    control_msgs::FollowJointTrajectoryGoal move_to_angle_goal;

    bool new_angles;
    bool cancel_move_to_angles_pending;
    //Variables
    /**
      * \brief
      *
      */
    iri_joints_module_state_t state;
    /**
      * \brief
      *
      */
    iri_joints_module_status_t status;

  protected:
    /**
      * \brief
      *
      */
    void state_machine(void);
    /**
      * \brief
      *
      */
    void reconfigure_callback(iri_joints_module::IriJointsModuleConfig &config, uint32_t level);

  public:
    /**
      * \brief Constructor
      *
      */
    CIriJointsModule(const std::string &name,const std::string &name_space=std::string(""));

    /*PointTo action*/
    /**
      * \brief Point to a point (x,y,z) at a desired maximum velocity.
      *
      * This function try to get the inverse kinematics for this point and
      * synchronizes the movement of the servos to arrive at the same time.
      *
      * \param x X coordenate on point_frame.
      *
      * \param y Y coordenate on point_frame.
      *
      * \param z Z coordenate on point_frame.
      *
      * \param max_vel The maximum velocity for the servos.
      *
      * \return True if action is called.
      */
    bool point_to(double x, double y, double z, double max_vel);

    /**
      * \brief Function to cancel point_to action.
      *
      */
    void cancel_point_to(void);

    /**
      * \brief Function to know if point to action is finished.
      *
      * \return True if finished.
      */
    bool is_point_to_finished(void);

    /*Tracker action*/
    /**
      * \brief Update the target at x,y,z position.
      *
      * \param x X coordenate on point_frame.
      *
      * \param y Y coordenate on point_frame.
      *
      * \param z Z coordenate on point_frame.
      *
      * \return True if action is called.
      */
    bool target_update(double x, double y, double z);

    /**
      * \brief Update the target at x,y,z position.
      *
      * \param point cartesian position to point to, with a valid reference frame id.
      *
      * \return True if action is called.
      */
    bool target_update(geometry_msgs::PointStamped &point);

    /**
      * \brief Function to start the tracker action.
      *
      * \return True if action is called.
      *
      */
    bool enable_tracker(void);

    /**
      * \brief Function to cancel tracking action.
      *
      */
    void cancel_tracker(void);

    /**
      * \brief Function to know if tracking action is finished.
      *
      * \return True if finished.
      */
    bool is_tracker_finished(void);

    /*Move to angles*/
    /**
      * \brief Function to move the joints to a desired angles at a desired velocities.
      *
      * \param joint_names String with the names ogf each joint.
      *
      * \param angles Desired angle for each joint in the same order than in joint_names.
      *
      * \param vel Cesired velocities for each joint in the same order than in joint_names.
      *
      * \return True if action is called.
      */
    bool move_to_angles(std::vector<std::string> joint_names, std::vector<double> angles, std::vector<double> vel);

    /**
      * \brief Function to cancel move to angles action.
      *
      */
    void cancel_move_to_angles(void);

    /**
      * \brief Function to know if move to angles action is finished.
      *
      * \return True if finished.
      */
    bool is_move_to_angles_finished(void);

    /*Common*/
    /**
      * \brief Function to change the inverse kinematics solver tolerance.
      *
      * \param tolerance Inverse kinematics tolerance.
      *
      * \return True if succeeded.
      */
    bool set_tolerance(double tolerance);

    /**
      * \brief Function to change the kinematic chain base frame
      *
      * \param tolerance The base frame.
      *
      * \return True if succeeded.
      */
    bool set_base_frame(std::string &base_frame);

    /**
      * \brief Function to change the ROS param name to get the urdf
      *
      * \param tolerance The ROS param.
      *
      * \return True if succeeded.
      */
    bool set_urdf_param(std::string &urdf_param);

    //Change?; Problem dynamic reconfigure module, solution param out of DynRec
    //void set_point_frame(std::string &point_frame);
    //Change?; Problem dynamic reconfigure module, solution param out of DynRec
    //void set_pointing_frame(std::string &pointing_frame);
    //Change?; Problem dynamic reconfigure module, solution param out of DynRec
    //void set_pointing_axis(uint8_t axis);

    //All the point_to param; Change?
    // void set_max_ik_tol(double max_ik_tol);
    // void set_max_ik_attemps(uint max_ik_attemps);
    // void set_max_fk_tol(double max_fk_tol);
    // void set_max_fk_attemps(uint max_fk_attemps);
    // void set_extra_rot_limits(double extra_rot_limits);
    // void set_point_to_header_frame_id(std::string &base_frame);

    /**
      * \brief Returns the current state
      *
      * This function returns the current status of the navigation module. It returns
      * one of the iri_joints_module_status_t values. This can be called at any time, but is
      * has only meaning when a goal has just finished.
      *
      * \return the current status of the module.
      */
    iri_joints_module_status_t get_status(void);
    /**
      * \brief Sets a parameter of type boolean
      *
      * This functions sets a parameter of type boolean. It calls the dynamic
      * reconfigure server with the provided data and waits for an answer. If
      * the data has not been changed, this function will report an error, as
      * well as if the service call itself fails or is pending.
      *
      * \param name String with the name of the parameter to change. This must
      *             coincide with a parameter name on the associated dynamic
      *             reconfigure server.
      *
      * \param value A boolean with the new value for the parameter.
      *
      * \return A boolean indicating whether the parameter has been successfully
      *         changed (true) or not (false)
      */
    bool set_parameter(const std::string &name,bool value);
    /**
      * \brief Sets a parameter of type integer
      *
      * This functions sets a parameter of type integer. It calls the dynamic
      * reconfigure server with the provided data and waits for an answer. If
      * the data has not been changed, this function will report an error, as
      * well as if the service call itself fails or is pending.
      *
      * \param name String with the name of the parameter to change. This must
      *             coincide with a parameter name on the associated dynamic
      *             reconfigure server.
      *
      * \param value An integer with the new value for the parameter. This value
      *              must be inside the valid value range of the parameter.
      *
      * \return A boolean indicating whether the parameter has been successfully
      *         changed (true) or not (false)
      */
    bool set_parameter(const std::string &name,int value);
    /**
      * \brief Sets a parameter of type string
      *
      * This functions sets a parameter of type string. It calls the dynamic
      * reconfigure server with the provided data and waits for an answer. If
      * the data has not been changed, this function will report an error, as
      * well as if the service call itself fails or is pending.
      *
      * \param name String with the name of the parameter to change. This must
      *             coincide with a parameter name on the associated dynamic
      *             reconfigure server.
      *
      * \param value A string with the new value for the parameter. This value
      *              must be inside the valid value range of the parameter.
      *
      * \return A boolean indicating whether the parameter has been successfully
      *         changed (true) or not (false)
      */
    bool set_parameter(const std::string &name,std::string &value);
    /**
      * \brief Sets a parameter of type double
      *
      * This functions sets a parameter of type double. It calls the dynamic
      * reconfigure server with the provided data and waits for an answer. If
      * the data has not been changed, this function will report an error, as
      * well as if the service call itself fails or is pending.
      *
      * \param name String with the name of the parameter to change. This must
      *             coincide with a parameter name on the associated dynamic
      *             reconfigure server.
      *
      * \param value A double with the new value for the parameter. This value
      *              must be inside the valid value range of the parameter.
      *
      * \return A boolean indicating whether the parameter has been successfully
      *         changed (true) or not (false)
      */
    bool set_parameter(const std::string &name,double value);

    /**
      * \brief Destructor
      *
      */
    ~CIriJointsModule();
  };

  #endif
